from .fundamental_component import FundamentalComponent

__all__ = [
    "DebuggerConnector",
]


class DebuggerConnector(FundamentalComponent):
    pass


DebuggerConnector._unique_id = "DebuggerConnector-builtin"
